@@ -83,6 +83,13 @@ static void panel_widget_end_move           (PanelWidget      *panel);
 static gboolean panel_widget_real_focus     (GtkWidget        *widget,
                                              GtkDirectionType  direction);
 
+static gboolean panel_widget_push_applet_right (PanelWidget *panel,
+						GList       *list,
+						int          push);
+static gboolean panel_widget_push_applet_left  (PanelWidget *panel,
+						GList       *list,
+						int          push);
+
 /************************
  debugging
  ************************/
@@ -120,6 +127,15 @@ applet_data_compare (AppletData *ad1, AppletData *ad2)
 	return ad1->pos - ad2->pos;
 }
 
+static void
+emit_applet_moved (PanelWidget *panel_widget,
+		   AppletData  *applet)
+{
+	g_signal_emit (panel_widget,
+		       panel_widget_signals [APPLET_MOVE_SIGNAL], 0,
+		       applet->applet);
+}
+
 /************************
  widget core
  ************************/
@@ -557,7 +573,7 @@ panel_widget_cadd (GtkContainer *container,
 	g_return_if_fail (PANEL_IS_WIDGET (container));
 	g_return_if_fail (GTK_IS_WIDGET (widget));
 
-	panel_widget_add (PANEL_WIDGET (container), widget, 0, FALSE);
+	panel_widget_add (PANEL_WIDGET (container), widget, FALSE, 0, FALSE);
 
 	p = g_object_get_data (G_OBJECT(widget),
 			       PANEL_APPLET_ASSOC_PANEL_KEY);
@@ -593,11 +609,8 @@ panel_widget_cremove (GtkContainer *container, GtkWidget *widget)
 	if (GTK_CONTAINER_CLASS (panel_widget_parent_class)->remove)
 		(* GTK_CONTAINER_CLASS (panel_widget_parent_class)->remove) (container,
 								widget);
-	if (ad != NULL) {
-		PanelWidget *panel = PANEL_WIDGET (container);
-
+	if (ad)
 		panel->applet_list = g_list_remove (panel->applet_list, ad);
-	}
 
 	g_signal_emit (G_OBJECT (container),
 		       panel_widget_signals[APPLET_REMOVED_SIGNAL],
@@ -630,32 +643,33 @@ get_applet_list_pos (PanelWidget *panel,
 
 /*tells us if an applet is "stuck" on the right side*/
 int
-panel_widget_is_applet_stuck(PanelWidget *panel, GtkWidget *applet)
+panel_widget_is_applet_stuck (PanelWidget *panel_widget,
+			      GtkWidget   *widget)
 {
-	AppletData *ad;
-	GList *list;
-	int i;
+	AppletData *applet;
 
-	g_return_val_if_fail(PANEL_IS_WIDGET(panel),FALSE);
-	g_return_val_if_fail(GTK_IS_WIDGET(applet),FALSE);
+	g_return_val_if_fail (PANEL_IS_WIDGET (panel_widget), FALSE);
+	g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
 
-	ad = g_object_get_data (G_OBJECT(applet), PANEL_APPLET_DATA);
+	applet = g_object_get_data (G_OBJECT (widget), PANEL_APPLET_DATA);
+	if (applet) {
+		GList *applet_list, *l;
+		int    end_pos = -1;
+	
+		applet_list = g_list_find (panel_widget->applet_list, applet);
 
-	g_return_val_if_fail(ad!=NULL,FALSE);
+		for (l = applet_list; l; l = l->next) {
+			applet = l->data;
 
-	list = g_list_find(panel->applet_list,ad);
+			if (end_pos != -1 && applet->pos != end_pos)
+				break;
 
-	g_return_val_if_fail(list!=NULL,FALSE);
+			end_pos = applet->pos + applet->cells;
+			if (end_pos >= panel_widget->size)
+				return TRUE;
+		}
+	}
 
-	do {
-		i=ad->pos+ad->cells;
-		if(i==panel->size)
-			return TRUE;
-		list = g_list_next(list);
-		if(!list)
-			break;
-		ad = list->data;
-	} while(ad->pos==i);
 	return FALSE;
 }
 
@@ -677,7 +691,7 @@ get_size_from_hints (AppletData *ad, int cells)
 		}
 	}
 	
-	return cells;
+	return MAX (cells, ad->min_cells);
 }
 
 static int
@@ -709,10 +723,10 @@ allocate_dirty_child(gpointer data)
 		if (panel->packed && ad->expand_major && ad->size_hints)
 			challoc.width = CLAMP (MAX (ad->size_hints [0], chreq.width),
 					       chreq.width,
-					       GTK_WIDGET (panel)->allocation.width - ad->pos);
+					       GTK_WIDGET (panel)->allocation.width - ad->constrained);
 
 		ad->cells = challoc.width; 
-		challoc.x = ad->pos;
+		challoc.x = ad->constrained;
 		challoc.y = GTK_WIDGET(panel)->allocation.height / 2 -
 			     challoc.height / 2;
 	} else {
@@ -722,51 +736,40 @@ allocate_dirty_child(gpointer data)
 		if (panel->packed && ad->expand_major && ad->size_hints)
 			challoc.height = CLAMP (MAX (ad->size_hints [0], chreq.height),
 						chreq.height,
-						GTK_WIDGET (panel)->allocation.height - ad->pos);
+						GTK_WIDGET (panel)->allocation.height - ad->constrained);
 
 		ad->cells = challoc.height;
 		challoc.x = GTK_WIDGET(panel)->allocation.width / 2 -
 			    challoc.width / 2;
-		challoc.y = ad->pos;
+		challoc.y = ad->constrained;
 	}
 
 	ad->min_cells = ad->cells;
 
 	if (!panel->packed && ad->expand_major) {
 		GList *l = g_list_find (panel->applet_list, ad);
-		int cells;
 		
 		if (l) {
 			if (l->next) {
 				nad = l->next->data;
-				cells = nad->pos - ad->pos - 1;
+				ad->cells = nad->constrained - ad->constrained - 1;
 			} else {
-				cells = panel->size - ad->pos - 1;
+				ad->cells = panel->size - ad->constrained - 1;
 			}
 
 			if (ad->size_hints) 
-				cells = get_size_from_hints (ad, cells);
-			
-			cells = MAX (cells, ad->min_cells);
-			
-			ad->cells = cells;
+				ad->cells = get_size_from_hints (ad, ad->cells);
 			
 			if (panel->orient == GTK_ORIENTATION_HORIZONTAL) {
 				challoc.width = ad->cells;
 			} else {
 				challoc.height = ad->cells;
 			}
-				
-
 		}
 	}
 
 	gtk_widget_size_allocate(ad->applet,&challoc);
 
-	g_signal_emit (G_OBJECT(panel),
-		       panel_widget_signals[APPLET_MOVE_SIGNAL],
-		       0, ad->applet);
-
 	return FALSE;
 }
 
@@ -777,6 +780,40 @@ panel_widget_queue_applet_for_resize(AppletData *ad)
 	g_idle_add(allocate_dirty_child,ad);
 }
 
+static void
+panel_widget_jump_applet_right (PanelWidget *panel,
+				GList       *list,
+				GList       *next,
+				int          pos)
+{
+	AppletData *ad;
+	AppletData *nad = NULL;
+
+	ad = list->data;
+	if (next)
+		nad = next->data;
+
+	if (pos >= panel->size)
+		return;
+
+	if (!nad || nad->constrained >= pos + ad->min_cells)
+		goto jump_right;
+
+	if (!panel_widget_push_applet_right (panel, next, pos + ad->min_cells - nad->constrained)) {
+		panel_widget_jump_applet_right (panel,
+						list,
+						next->next,
+						nad->constrained + nad->min_cells);
+		return;
+	}
+
+ jump_right:
+	ad->pos = ad->constrained = pos;
+	panel->applet_list = g_list_remove_link (panel->applet_list, list);
+	panel->applet_list = panel_g_list_insert_before (panel->applet_list, next, list);
+	panel_widget_queue_applet_for_resize (ad);
+	emit_applet_moved (panel, ad);
+}
 
 static void
 panel_widget_switch_applet_right (PanelWidget *panel,
@@ -784,30 +821,77 @@ panel_widget_switch_applet_right (PanelWidget *panel,
 {
 	AppletData *ad;
 	AppletData *nad = NULL;
-
-	g_return_if_fail (PANEL_IS_WIDGET (panel));
-	g_return_if_fail (list != NULL);
 	
-	ad = list->data;
+	g_assert (list != NULL);
 
-	if (ad->pos + ad->min_cells >= panel->size)
+	ad = list->data;
+	if (ad->constrained + ad->min_cells >= panel->size)
 		return;
 
 	if (list->next)
 		nad = list->next->data;
 
-	if (!nad || nad->pos > ad->pos + ad->min_cells) {
-		ad->pos += MOVE_INCREMENT;
+	if (!nad || nad->constrained >= ad->constrained + ad->min_cells + MOVE_INCREMENT) {
+		ad->pos = ad->constrained += MOVE_INCREMENT;
 		panel_widget_queue_applet_for_resize (ad);
+		emit_applet_moved (panel, ad);
 		return;
 	}
 
-	nad->pos = ad->pos;
-	ad->pos = nad->pos + nad->min_cells;
-	panel->applet_list = panel_g_list_swap_next (panel->applet_list,list);
+	if (nad->locked) {
+		panel_widget_jump_applet_right (panel,
+						list,
+						list->next->next,
+						nad->constrained + nad->min_cells);
+		return;
+	}
+						  
+	nad->constrained = nad->pos = ad->constrained;
+	ad->constrained = ad->pos = ad->constrained + nad->min_cells;
+	panel->applet_list = panel_g_list_swap_next (panel->applet_list, list);
 
 	panel_widget_queue_applet_for_resize (ad);
 	panel_widget_queue_applet_for_resize (nad);
+
+	emit_applet_moved (panel, ad);
+	emit_applet_moved (panel, nad);
+}
+
+static void
+panel_widget_jump_applet_left (PanelWidget *panel,
+			       GList       *list,
+			       GList       *prev,
+			       int          pos)
+{
+	AppletData *ad;
+	AppletData *pad = NULL;
+
+	ad = list->data;
+	if (prev)
+		pad = prev->data;
+
+	if (pos < 0)
+		return;
+
+	if (!pad || pad->constrained + pad->min_cells <= pos)
+		goto jump_left;
+
+	if (!panel_widget_push_applet_left (panel, prev, pad->constrained + pad->min_cells - pos)) {
+		panel_widget_jump_applet_left (panel,
+					       list,
+					       prev->prev,
+					       pad->constrained - ad->min_cells);
+		return;
+	}
+
+ jump_left:
+	ad->pos = ad->constrained = pos;
+	panel->applet_list = g_list_remove_link (panel->applet_list, list);
+	panel->applet_list = panel_g_list_insert_after (panel->applet_list, prev, list);
+	panel_widget_queue_applet_for_resize (ad);
+	if (pad && (pad->constrained + pad->cells) >= ad->constrained)
+		panel_widget_queue_applet_for_resize (pad);
+	emit_applet_moved (panel, ad);
 }
 
 static void
@@ -817,81 +901,197 @@ panel_widget_switch_applet_left (PanelWidget *panel,
 	AppletData *ad;
 	AppletData *pad = NULL;
 
-	g_return_if_fail (PANEL_IS_WIDGET (panel));
-	g_return_if_fail (list != NULL);
-	
 	ad = list->data;
-
-	if (ad->pos <= 0)
+	if (ad->constrained <= 0)
 		return;
 
 	if (list->prev)
 		pad = list->prev->data;
 
-	if (!pad || pad->pos + pad->min_cells < ad->pos) {
-		ad->pos -= MOVE_INCREMENT;
+	if (!pad || pad->constrained + pad->min_cells <= ad->constrained - MOVE_INCREMENT) {
+		ad->pos = ad->constrained -= MOVE_INCREMENT;
 		panel_widget_queue_applet_for_resize (ad);
-		if (pad && (pad->pos + pad->cells) >= ad->pos)
+		if (pad && (pad->constrained + pad->cells) >= ad->constrained)
 			panel_widget_queue_applet_for_resize (pad);
+		emit_applet_moved (panel, ad);
 		return;
 	}
 
-	ad->pos = pad->pos;
-	pad->pos = ad->pos + ad->min_cells;
+	if (pad->locked) {
+		panel_widget_jump_applet_left (panel,
+					       list,
+					       list->prev->prev,
+					       pad->constrained - ad->min_cells);
+		return;
+	}
+
+	ad->constrained = ad->pos = pad->constrained;
+	pad->constrained = pad->pos = ad->constrained + ad->min_cells;
 	panel->applet_list = panel_g_list_swap_prev (panel->applet_list, list);
 
 	panel_widget_queue_applet_for_resize (ad);
 	panel_widget_queue_applet_for_resize (pad);
+
+	emit_applet_moved (panel, ad);
+	emit_applet_moved (panel, pad);
 }
 
-static int
-panel_widget_get_right_switch_pos(PanelWidget *panel, GList *list)
+static gboolean
+panel_widget_try_push_right (PanelWidget *panel,
+			     GList       *list,
+			     int          push)
 {
 	AppletData *ad;
 	AppletData *nad = NULL;
 
-	g_return_val_if_fail(PANEL_IS_WIDGET(panel),-1);
-	g_return_val_if_fail(list!=NULL,-1);
-	
+	g_assert (list != NULL);
+
 	ad = list->data;
-
-
-	if(list->next)
+	if (list->next)
 		nad = list->next->data;
 
-	if(!nad || nad->pos > ad->pos+ad->cells) {
-		return ad->pos+1;
-	}
+	if (ad->locked)
+		return FALSE;
 
-	return nad->pos+nad->min_cells-ad->cells;
+	if (ad->constrained + ad->min_cells + push >= panel->size)
+		return FALSE;
+
+	if (!nad || nad->constrained >= ad->constrained + ad->min_cells + push)
+		return TRUE;
+
+	return panel_widget_try_push_right (panel, list->next, push);
 }
 
 static int
-panel_widget_get_left_switch_pos(PanelWidget *panel, GList *list)
+panel_widget_get_right_jump_pos (PanelWidget *panel,
+				 AppletData  *ad,
+				 GList       *next,
+				 int          pos)
+{
+	AppletData *nad = NULL;
+
+	if (next)
+		nad = next->data;
+
+	if (!nad || nad->constrained >= pos + ad->min_cells)
+		return pos;
+
+	if (panel_widget_try_push_right (panel, next, pos + ad->min_cells - nad->constrained))
+		return pos;
+
+	return panel_widget_get_right_jump_pos (panel,
+						ad,
+						next->next,
+						nad->constrained + nad->min_cells);
+}
+
+static int
+panel_widget_get_right_switch_pos (PanelWidget *panel,
+				   GList       *list)
+{
+	AppletData *ad;
+	AppletData *nad = NULL;
+
+	g_assert (list != NULL);
+
+	ad = list->data;
+	if (list->next)
+		nad = list->next->data;
+
+	if (!nad || nad->constrained >= ad->constrained + ad->min_cells + MOVE_INCREMENT)
+		return ad->constrained + MOVE_INCREMENT;
+
+	if (nad->locked)
+		return panel_widget_get_right_jump_pos (panel,
+							ad,
+							list->next->next,
+							nad->constrained + nad->min_cells);
+
+	return nad->constrained + nad->min_cells - ad->cells;
+}
+
+static gboolean
+panel_widget_try_push_left (PanelWidget *panel,
+			    GList       *list,
+			    int          push)
 {
 	AppletData *ad;
 	AppletData *pad = NULL;
 
-	g_return_val_if_fail(PANEL_IS_WIDGET(panel),-1);
-	g_return_val_if_fail(list!=NULL,-1);
-	
-	ad = list->data;
+	g_assert (list != NULL);
 
-	if(list->prev)
+	ad = list->data;
+	if (list->prev)
 		pad = list->prev->data;
-	if(!pad || pad->pos+pad->min_cells < ad->pos)
-		return ad->pos-1;
-	return pad->pos;
+
+	if (ad->locked)
+		return FALSE;
+
+	if (ad->constrained - push < 0)
+		return FALSE;
+	
+	if (!pad || pad->constrained + pad->min_cells <= ad->constrained - push)
+		return TRUE;
+
+	return panel_widget_try_push_left (panel, list->prev, push);
 }
 
+static int
+panel_widget_get_left_jump_pos (PanelWidget *panel,
+				AppletData  *ad,
+				GList       *prev,
+				int          pos)
+{
+	AppletData *pad = NULL;
+
+	if (prev)
+		pad = prev->data;
+
+	if (!pad || pad->constrained + pad->min_cells <= pos)
+		return pos;
+
+	if (panel_widget_try_push_left (panel, prev, pad->constrained + pad->min_cells - pos))
+		return pos;
+
+	return panel_widget_get_left_jump_pos (panel,
+					       ad,
+					       prev->prev,
+					       pad->constrained - ad->min_cells);
+}
+
+static int
+panel_widget_get_left_switch_pos (PanelWidget *panel,
+				  GList       *list)
+{
+	AppletData *ad;
+	AppletData *pad = NULL;
+
+	g_assert (list != NULL);
+
+	ad = list->data;
+	if (list->prev)
+		pad = list->prev->data;
+	
+	if (!pad || pad->constrained + pad->min_cells <= ad->constrained - MOVE_INCREMENT)
+		return ad->constrained - MOVE_INCREMENT;
+
+	if (pad->locked)
+		return panel_widget_get_left_jump_pos (panel,
+						       ad,
+						       list->prev->prev,
+						       pad->constrained - ad->min_cells);
+	
+	return pad->constrained;
+}
 
 static void
-panel_widget_switch_move (PanelWidget *panel, AppletData *ad, int moveby)
+panel_widget_switch_move (PanelWidget *panel,
+			  AppletData  *ad,
+			  int          moveby)
 {
-	int finalpos;
-	int pos;
 	GList *list;
-	AppletData *pad;
+	int    finalpos;
+	int    pos;
 
 	g_return_if_fail (ad != NULL);
 	g_return_if_fail (PANEL_IS_WIDGET (panel));
@@ -902,103 +1102,135 @@ panel_widget_switch_move (PanelWidget *panel, AppletData *ad, int moveby)
 	list = g_list_find (panel->applet_list, ad);
 	g_return_if_fail (list != NULL);
 
-	finalpos = ad->pos + moveby;
+	finalpos = ad->constrained + moveby;
+
+	if (ad->constrained < finalpos) {
+		AppletData *pad;
 
-	if (ad->pos < finalpos) {
-		while (ad->pos < finalpos) {
-			pos = panel_widget_get_right_switch_pos (panel, list);
-			if (pos + ad->min_cells > panel->size ||
-			    abs (pos - finalpos) >= abs (ad->pos - finalpos))
-				break;
-			panel_widget_switch_applet_right (panel, list);
-		}
 		if (list->prev) {
 			pad = list->prev->data;
 			if (pad->expand_major)
-				panel_widget_queue_applet_for_resize(pad);
+				panel_widget_queue_applet_for_resize (pad);
+		}
+
+		while (ad->constrained < finalpos) {
+			pos = panel_widget_get_right_switch_pos (panel, list);
+
+			if (abs (pos - finalpos) >= abs (ad->constrained - finalpos) ||
+			    pos + ad->min_cells > panel->size)
+				break;
+
+			panel_widget_switch_applet_right (panel, list);
+		}
+
+		if (list->prev) {
+			pad = list->prev->data;
+			if (pad->expand_major)
+				panel_widget_queue_applet_for_resize (pad);
 		}
 	} else {
-		while (ad->pos > finalpos) {
+		AppletData *nad;
+
+		if (list->next) {
+			nad = list->next->data;
+			if (nad->expand_major)
+				panel_widget_queue_applet_for_resize (nad);
+		}
+
+		while (ad->constrained > finalpos) {
 			pos = panel_widget_get_left_switch_pos (panel, list);
-			if (pos < 0 || abs (pos - finalpos) >= abs (ad->pos - finalpos))
+
+			if (abs (pos - finalpos) >= abs (ad->constrained - finalpos) || pos < 0)
 				break;
+
 			panel_widget_switch_applet_left (panel, list);
 		}
+		
 	}
 }
 
 static int
-push_applet_right(PanelWidget *panel, GList *list)
+panel_widget_push_applet_right (PanelWidget *panel,
+				GList       *list,
+				int          push)
 {
-	AppletData *ad = NULL;
+	AppletData *ad;
 	AppletData *nad = NULL;
 
-	g_return_val_if_fail(PANEL_IS_WIDGET(panel),FALSE);
+	g_assert (list != NULL);
 	
-	if (list) {
-		ad = list->data;
-	
-		if(ad->pos + ad->min_cells >= panel->size)
-			return FALSE;
+	ad = list->data;
+	if (ad->constrained + ad->min_cells + push >= panel->size)
+		return FALSE;
 
-		if(list->next)
-			nad = list->next->data;
-	}
+	if (ad->locked)
+		return FALSE;
 
-	if(!nad || nad->pos > ad->pos+ad->min_cells) {
-		ad->pos++;
-		panel_widget_queue_applet_for_resize(ad);
+	if (list->next)
+		nad = list->next->data;
+
+	if (!nad || nad->constrained >= ad->constrained + ad->min_cells + push) {
+		ad->pos = ad->constrained += push;
+		panel_widget_queue_applet_for_resize (ad);
+		emit_applet_moved (panel, ad);
 		return TRUE;
 	}
 
-	g_return_val_if_fail(list!=NULL,FALSE);
-	
-	if(!push_applet_right(panel,list->next)) {
+	g_assert (list->next != NULL);
+
+	if (!panel_widget_push_applet_right (panel, list->next, push))
 		return FALSE;
-	}
-	ad->pos++;
-	panel_widget_queue_applet_for_resize(ad);
+
+	ad->pos = ad->constrained += push;;
+	panel_widget_queue_applet_for_resize (ad);
+	emit_applet_moved (panel, ad);
+	
 	return TRUE;
 }
 
 static int
-push_applet_left(PanelWidget *panel, GList *list)
+panel_widget_push_applet_left (PanelWidget *panel,
+			       GList       *list,
+			       int          push)
 {
-	AppletData *ad = NULL;
+	AppletData *ad;
 	AppletData *pad = NULL;
 
-	g_return_val_if_fail(PANEL_IS_WIDGET(panel),FALSE);
+	g_assert (list != NULL);
 
-	if (list) {
-		ad = list->data;
+	ad = list->data;
+	if (ad->constrained - push < 0)
+		return FALSE;
 
-		if(ad->pos <= 0)
-			return FALSE;
+	if (ad->locked)
+		return FALSE;
 
-		if(list->prev)
-			pad = list->prev->data;
-	}
+	if (list->prev)
+		pad = list->prev->data;
 
-	if(!list || !pad || pad->pos+pad->min_cells < ad->pos) {
-		ad->pos--;
-		panel_widget_queue_applet_for_resize(ad);
-		if (pad && pad->pos+pad->cells >= ad->pos)
-			panel_widget_queue_applet_for_resize(pad);
+	if (!pad || pad->constrained + pad->min_cells <= ad->constrained - push) {
+		ad->pos = ad->constrained -= push;
+		panel_widget_queue_applet_for_resize (ad);
+		emit_applet_moved (panel, ad);
 		return TRUE;
 	}
 
-	g_return_val_if_fail(list!=NULL,FALSE);
-	
-	if(!push_applet_left(panel,list->prev)) {
+	g_assert (list->prev != NULL);
+
+	if (!panel_widget_push_applet_left (panel, list->prev, push))
 		return FALSE;
-	}
-	ad->pos--;
-	panel_widget_queue_applet_for_resize(ad);
+
+	ad->pos = ad->constrained -= push;
+	panel_widget_queue_applet_for_resize (ad);
+	emit_applet_moved (panel, ad);
+	
 	return TRUE;
 }
 
 static void
-panel_widget_push_move (PanelWidget *panel, AppletData *ad, int moveby)
+panel_widget_push_move (PanelWidget *panel,
+			AppletData  *ad,
+			int          moveby)
 {
 	AppletData *pad;
 	int finalpos;
@@ -1013,23 +1245,22 @@ panel_widget_push_move (PanelWidget *panel, AppletData *ad, int moveby)
 	list = g_list_find (panel->applet_list, ad);
 	g_return_if_fail (list != NULL);
 
-	finalpos = ad->pos + moveby;
+	finalpos = ad->constrained + moveby;
 
-	if (ad->pos < finalpos) {
-		while (ad->pos < finalpos) {
-			if ( ! push_applet_right (panel, list))
+	if (ad->constrained < finalpos) {
+		while (ad->constrained < finalpos)
+			if (!panel_widget_push_applet_right (panel, list, 1))
 				break;
-		}
+
                 if (list->prev) {
 			pad = list->prev->data;
 			if (pad->expand_major)
 				panel_widget_queue_applet_for_resize(pad);
 		}
 	} else {
-                while (ad->pos > finalpos) {
-			if ( ! push_applet_left (panel, list))
+                while (ad->constrained > finalpos)
+			if (!panel_widget_push_applet_left (panel, list, 1))
 				break;
-		}
 	}
 }
 
@@ -1060,20 +1291,18 @@ panel_widget_right_stick(PanelWidget *panel,int old_size)
 
 	ad = list->data;
 	do { 
-		i = ad->pos;
-		ad->pos = pos--;
+		i = ad->constrained;
+		ad->pos = ad->constrained = pos--;
 		ad->cells = 1;
 		prev = list;
 		list = g_list_previous(list);
 		if(!list)
 			break;
 		ad = list->data;
-	} while(ad->pos+ad->cells == i);
+	} while(ad->constrained + ad->cells == i);
 
-	for(list = prev;list!=NULL;list=g_list_next(list))
-		g_signal_emit (G_OBJECT(panel),
-			       panel_widget_signals[APPLET_MOVE_SIGNAL],
-			       0, ((AppletData *)list->data)->applet);
+	for (list = prev; list; list = list->next)
+		emit_applet_moved (panel, list->data);
 }
 
 static void
@@ -1177,8 +1406,6 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 {
 	PanelWidget *panel;
 	GList *list;
-	GSList *send_move = NULL;
-	GSList *li;
 	int i;
 	int old_size;
 	int old_thick;
@@ -1218,11 +1445,9 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 			GtkAllocation challoc;
 			GtkRequisition chreq;
 			gtk_widget_get_child_requisition(ad->applet,&chreq);
+
+			ad->constrained = i;
 			
-			if(i != ad->pos) {
-				ad->pos = i;
-				send_move = g_slist_prepend(send_move,ad);
-			}
 			challoc.width = chreq.width;
 			challoc.height = chreq.height;
 			if(panel->orient == GTK_ORIENTATION_HORIZONTAL) {
@@ -1235,7 +1460,7 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 							       allocation->width - i);
 
 				ad->cells = challoc.width;
-				challoc.x = ad->pos;
+				challoc.x = ad->constrained;
 				challoc.y = allocation->height / 2 - challoc.height / 2;
 			} else {
 				if (ad->expand_minor)
@@ -1248,7 +1473,7 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 
 				ad->cells = challoc.height;
 				challoc.x = allocation->width / 2 - challoc.width / 2;
-				challoc.y = ad->pos;
+				challoc.y = ad->constrained;
 			}
 			ad->min_cells  = ad->cells;
 			ad->dirty = FALSE;
@@ -1272,9 +1497,11 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 				ad->cells = chreq.height;
 
 			ad->min_cells = ad->cells;
+
+			ad->constrained = ad->pos;
 			
 			if (ad->expand_major) {
-				cells = (i - ad->pos) - 1;
+				cells = (i - ad->constrained) - 1;
 
 				if (ad->size_hints) 
 					cells = get_size_from_hints (ad, cells);
@@ -1283,11 +1510,10 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 				ad->cells = cells;
 			}
 
-			if(ad->pos+ad->min_cells > i) {
-				ad->pos = i - ad->min_cells;
-				send_move = g_slist_prepend(send_move,ad);
-			}
-			i = ad->pos;
+			if (ad->constrained + ad->min_cells > i)
+				ad->constrained = i - ad->min_cells;
+
+			i = ad->constrained;
 		}
 		if(i < 0) {
 			i = 0;
@@ -1296,13 +1522,10 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 			    list = g_list_next(list)) {
 				AppletData *ad = list->data;
 				
-				if(ad->pos < i) {
-					ad->pos = i;
-					if(!g_slist_find(send_move,ad))
-						send_move = g_slist_prepend(send_move,ad);
-				}
+				if (ad->constrained < i)
+					ad->constrained = i;
 				
-				i = ad->pos + ad->cells;
+				i = ad->constrained + ad->cells;
 			}
 		}
 
@@ -1320,7 +1543,7 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 				if (ad->expand_minor) {
 					challoc.height = allocation->height;
 				}
-				challoc.x = ad->pos;
+				challoc.x = ad->constrained;
 				challoc.y = allocation->height / 2 - challoc.height / 2;
 			} else {
 				challoc.height = ad->cells;
@@ -1328,7 +1551,7 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 					challoc.width = allocation->width;
 				}
 				challoc.x = allocation->width / 2 - challoc.width / 2;
-				challoc.y = ad->pos;
+				challoc.y = ad->constrained;
 			}
 			ad->dirty = FALSE;
 			
@@ -1340,14 +1563,6 @@ panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
 	else
 		panel->thick = allocation->width;
 
-	for(li=send_move;li!=NULL;li=g_slist_next(li)) {
-		AppletData *ad = li->data;
-		g_signal_emit (G_OBJECT(panel),
-			       panel_widget_signals[APPLET_MOVE_SIGNAL],
-			       0, ad->applet);
-	}
-	g_slist_free(send_move);
-
 	panel_widget_set_background_region (panel);
 }
 
@@ -1528,18 +1743,21 @@ static guint moving_timeout = 0;
 static gboolean been_moved = FALSE;
 static gboolean repeat_if_outside = FALSE;
 
-void
+static gboolean
 panel_widget_applet_drag_start_no_grab (PanelWidget *panel,
 					GtkWidget *applet,
 					int drag_off)
 {
 	AppletData *ad;
 
-	g_return_if_fail (PANEL_IS_WIDGET (panel));
-	g_return_if_fail (GTK_IS_WIDGET (panel));
+	g_return_val_if_fail (PANEL_IS_WIDGET (panel), FALSE);
+	g_return_val_if_fail (GTK_IS_WIDGET (panel), FALSE);
 
 	ad = g_object_get_data (G_OBJECT (applet), PANEL_APPLET_DATA);
-	g_return_if_fail (ad != NULL);
+	g_return_val_if_fail (ad != NULL, FALSE);
+
+	if (ad->locked)
+		return FALSE;
 
 	if (moving_timeout != 0) {
 		g_source_remove (moving_timeout);
@@ -1553,7 +1771,7 @@ panel_widget_applet_drag_start_no_grab (PanelWidget *panel,
 #endif
 	panel->currently_dragged_applet = ad;
 	if (drag_off == PW_DRAG_OFF_CURSOR)
-		ad->drag_off = panel_widget_get_cursorloc (panel) - ad->pos;
+		ad->drag_off = panel_widget_get_cursorloc (panel) - ad->constrained;
 	else if (drag_off == PW_DRAG_OFF_CENTER)
 		ad->drag_off = ad->cells / 2;
 	else
@@ -1562,10 +1780,12 @@ panel_widget_applet_drag_start_no_grab (PanelWidget *panel,
 	add_all_move_bindings (panel);
 
 	panel_applet_in_drag = TRUE;
+	
+	return TRUE;
 }
 
 
-void
+static void
 panel_widget_applet_drag_end_no_grab (PanelWidget *panel)
 {
 	g_return_if_fail (panel != NULL);
@@ -1598,7 +1818,8 @@ panel_widget_applet_drag_start (PanelWidget *panel,
 	g_message("Starting drag [grabbed] on a %s at %p\n",
 		  g_type_name(G_TYPE_FROM_INSTANCE(applet)), applet);
 #endif
-	panel_widget_applet_drag_start_no_grab (panel, applet, drag_off);
+	if (!panel_widget_applet_drag_start_no_grab (panel, applet, drag_off))
+		return;
 
 	gtk_grab_add (applet);
 	if (applet->window) {
@@ -1665,16 +1886,16 @@ walk_up_to (int pos, GList *list)
 
 	ad = list->data;
 
-	if (ad->pos <= pos &&
-	    ad->pos + ad->cells > pos)
+	if (ad->constrained <= pos &&
+	    ad->constrained + ad->cells > pos)
 		return list;
 	while (list->next != NULL &&
-	       ad->pos + ad->cells <= pos) {
+	       ad->constrained + ad->cells <= pos) {
 		list = list->next;
 		ad = list->data;
 	}
 	while (list->prev != NULL &&
-	       ad->pos > pos) {
+	       ad->constrained > pos) {
 		list = list->prev;
 		ad = list->data;
 	}
@@ -1686,14 +1907,16 @@ is_in_applet (int pos, AppletData *ad)
 {
 	g_return_val_if_fail (ad != NULL, NULL);
 
-	if (ad->pos <= pos &&
-	    ad->pos + ad->min_cells > pos)
+	if (ad->constrained <= pos &&
+	    ad->constrained + ad->min_cells > pos)
 		return ad->applet;
 	return NULL;
 }
 
 static int
-panel_widget_get_free_spot (PanelWidget *panel, AppletData *ad, int place)
+panel_widget_get_free_spot (PanelWidget *panel,
+			    AppletData  *ad,
+			    int          place)
 {
 	int i, e;
 	int start;
@@ -1703,7 +1926,7 @@ panel_widget_get_free_spot (PanelWidget *panel, AppletData *ad, int place)
 	g_return_val_if_fail (PANEL_IS_WIDGET (panel), -1);
 	g_return_val_if_fail (ad != NULL, -1);
 
-	if (ad->pos >= panel->size)
+	if (ad->constrained >= panel->size)
 		return -1;
 
 	if (panel->applet_list == NULL) {
@@ -1769,19 +1992,19 @@ panel_widget_get_free_spot (PanelWidget *panel, AppletData *ad, int place)
 	}
 }
 
-/*to call this function we MUST know that there is at least
-  ad->cells free at pos otherwise we will mess up the panel*/
 static void
-panel_widget_nice_move (PanelWidget *panel, AppletData *ad, int pos)
+panel_widget_nice_move (PanelWidget *panel,
+			AppletData  *ad,
+			int          pos)
 {
 	AppletData *pad1 = NULL, *pad2 = NULL;
-	GList *l;
+	GList      *l;
 	
 	g_return_if_fail (PANEL_IS_WIDGET (panel));
 	g_return_if_fail (ad != NULL);
 
-	if (pos < 0 ||
-	    pos == ad->pos)
+	pos = panel_widget_get_free_spot (panel, ad, pos);
+	if (pos < 0 || pos == ad->pos)
 		return;
 
 	l = g_list_find (panel->applet_list, ad);
@@ -1789,7 +2012,7 @@ panel_widget_nice_move (PanelWidget *panel, AppletData *ad, int pos)
 		pad1 = l->prev->data;
 	}
 
-	ad->pos = pos;
+	ad->pos = ad->constrained = pos;
 
 	panel->applet_list =
 		panel_g_list_resort_item (panel->applet_list, ad,
@@ -1809,6 +2032,8 @@ panel_widget_nice_move (PanelWidget *panel, AppletData *ad, int pos)
 	if (pad2 && pad2 != pad1 && pad2->expand_major) {
 		panel_widget_queue_applet_for_resize (pad2);
 	}
+
+	emit_applet_moved (panel, ad);
 }
 
 /* schedule to run the below function */
@@ -1833,7 +2058,7 @@ panel_widget_applet_move_to_cursor (PanelWidget *panel)
 
 	ad = panel->currently_dragged_applet;
 
-	pos = ad->pos;
+	pos = ad->constrained;
 
 	applet = ad->applet;
 	g_assert(GTK_IS_WIDGET(applet));
@@ -1892,15 +2117,13 @@ panel_widget_applet_move_to_cursor (PanelWidget *panel)
 			movement = PANEL_FREE_MOVE;
 	}
 	
-	switch(movement) {
+	switch (movement) {
 	case PANEL_SWITCH_MOVE:
 		moveby = panel_widget_get_moveby (panel, pos, ad->drag_off);
 		panel_widget_switch_move (panel, ad, moveby);
 		break;
 	case PANEL_FREE_MOVE:
-		pos = panel_widget_get_free_spot (panel, ad,
-					panel_widget_get_cursorloc (panel));
-		panel_widget_nice_move (panel, ad, pos);
+		panel_widget_nice_move (panel, ad, panel_widget_get_cursorloc (panel));
 		break;
 	case PANEL_PUSH_MOVE:
 		moveby = panel_widget_get_moveby (panel, pos, ad->drag_off);
@@ -2140,11 +2363,13 @@ panel_widget_find_empty_pos(PanelWidget *panel, int pos)
 	GList *list;
 
 	g_return_val_if_fail(PANEL_IS_WIDGET(panel),-1);
-	g_return_val_if_fail(pos>=0,-1);
 
 	if(pos>=panel->size)
 		pos = panel->size-1;
 
+	if (pos <= 0)
+		pos = 0;
+
 	if(!panel->applet_list)
 		return pos;
 
@@ -2192,6 +2417,7 @@ panel_widget_add_forbidden (PanelWidget *panel)
 int
 panel_widget_add (PanelWidget *panel,
 		  GtkWidget   *applet,
+		  gboolean     locked,
 		  int          pos,
 		  gboolean     insert_at_pos)
 {
@@ -2199,17 +2425,13 @@ panel_widget_add (PanelWidget *panel,
 
 	g_return_val_if_fail (PANEL_IS_WIDGET (panel), -1);
 	g_return_val_if_fail (GTK_IS_WIDGET (applet), -1);
-	g_return_val_if_fail (pos >= 0, -1);
 
 	ad = g_object_get_data (G_OBJECT (applet), PANEL_APPLET_DATA);
 
 	if (ad != NULL)
 		pos = ad->pos;
 
-	if (pos < 0)
-		pos = 0;
-	
-	if ( ! insert_at_pos) {
+	if (!insert_at_pos || pos < 0) {
 		if (panel->packed) {
 			if (get_applet_list_pos (panel, pos)) 
 				/*this is a slight hack so that this applet
@@ -2236,11 +2458,13 @@ panel_widget_add (PanelWidget *panel,
 		ad->cells = 1;
 		ad->min_cells = 1;
 		ad->pos = pos;
+		ad->constrained = pos;
 		ad->drag_off = 0;
 		ad->dirty = FALSE;
 		ad->no_die = 0;
 		ad->expand_major = FALSE;
 		ad->expand_minor = FALSE;
+		ad->locked = locked;
 		ad->size_hints = NULL;
 		g_object_set_data (G_OBJECT (applet),
 				   PANEL_APPLET_DATA, ad);
@@ -2282,6 +2506,7 @@ panel_widget_reparent (PanelWidget *old_panel,
 {
 	AppletData *ad;
 	GtkWidget *focus_widget = NULL;
+	AppletInfo* info;
 
 	g_return_val_if_fail(PANEL_IS_WIDGET(old_panel),-1);
 	g_return_val_if_fail(PANEL_IS_WIDGET(new_panel),-1);
@@ -2297,10 +2522,13 @@ panel_widget_reparent (PanelWidget *old_panel,
 	if (panel_toplevel_get_is_hidden (new_panel->toplevel))
 		return FALSE;
 	
+	info = g_object_get_data (G_OBJECT (ad->applet), "applet_info");
+
 	/*we'll resize both panels anyway*/
 	ad->dirty = FALSE;
 	
-	ad->pos = pos;
+	ad->pos = ad->constrained = panel_widget_get_free_spot (new_panel, ad, pos);
+	panel_widget_queue_applet_for_resize (ad);
 
 	ad->no_die++;
 
@@ -2309,6 +2537,9 @@ panel_widget_reparent (PanelWidget *old_panel,
 		focus_widget = gtk_window_get_focus (GTK_WINDOW (old_panel->toplevel));
 	gtk_widget_reparent (applet, GTK_WIDGET (new_panel));
 
+	if (info && info->type == PANEL_OBJECT_BONOBO)
+		panel_applet_frame_set_panel (PANEL_APPLET_FRAME (ad->applet), new_panel);
+
 	if (GTK_WIDGET_CAN_FOCUS (new_panel))
 		GTK_WIDGET_UNSET_FLAGS (new_panel, GTK_CAN_FOCUS);
 	if (focus_widget) {
@@ -2326,6 +2557,8 @@ panel_widget_reparent (PanelWidget *old_panel,
 
 	ad->no_die--;
 
+	emit_applet_moved (new_panel, ad);
+
 	return TRUE;
 }
 
@@ -2422,9 +2655,8 @@ static void
 panel_widget_free_move_applet (PanelWidget      *panel,
                                GtkDirectionType  dir)
 {
-	AppletData       *ad;
-	gint increment = MOVE_INCREMENT;
-	gint pos;
+	AppletData *ad;
+	gint        increment = MOVE_INCREMENT;
 
 	ad = panel->currently_dragged_applet;
 
@@ -2442,32 +2674,30 @@ panel_widget_free_move_applet (PanelWidget      *panel,
 		return;
 	}
 
-	pos = panel_widget_get_free_spot (panel, ad, 
-                                          increment + ad->pos + ad->drag_off);
-	panel_widget_nice_move (panel, ad, pos);
+	panel_widget_nice_move (panel, ad, increment + ad->constrained + ad->drag_off);
 }
 
 static void
 panel_widget_tab_move (PanelWidget *panel,
                        gboolean     next)
 {
-	AppletData       *ad;
-	GSList *li;
 	PanelWidget *new_panel = NULL;
 	PanelWidget *previous_panel = NULL;
+	AppletData  *ad;
+	GSList      *l;
 
 	ad = panel->currently_dragged_applet;
 
 	if (!ad)
 		return;	
 	
-	for (li = panels; li; li = li->next) {
-		PanelWidget *panel_in_list = li->data;
+	for (l = panels; l; l = l->next) {
+		PanelWidget *panel_in_list = l->data;
 
 		if (panel_in_list == panel) {
 			if (next) {
-				if (li->next)
-					new_panel = li->next->data;
+				if (l->next)
+					new_panel = l->next->data;
 				else
 					new_panel = ((GSList *)panels)->data;
 				
@@ -2483,14 +2713,14 @@ panel_widget_tab_move (PanelWidget *panel,
 				previous_panel = panel_in_list;
 		}
 	}
-	g_return_if_fail (li);
-	if (!new_panel) {
-		if (previous_panel)
-			new_panel = previous_panel;
-	}
-	if (new_panel && (new_panel != panel)) {
+
+	g_return_if_fail (l != NULL);
+
+	if (!new_panel && previous_panel)
+		new_panel = previous_panel;
+	
+	if (new_panel && (new_panel != panel))
 		panel_widget_reparent (panel, new_panel, ad->applet, 0);
-	}
 }
 
 static void
@@ -2557,6 +2787,46 @@ panel_widget_set_applet_expandable (PanelWidget *panel,
 	gtk_widget_queue_resize (GTK_WIDGET (panel));
 }
 
+void
+panel_widget_set_applet_locked (PanelWidget *panel,
+				GtkWidget   *applet,
+				gboolean     locked)
+{
+	AppletData *ad;
+
+	ad = g_object_get_data (G_OBJECT (applet), PANEL_APPLET_DATA);
+	if (!ad)
+		return;
+
+	ad->locked = locked;
+}
+
+gboolean
+panel_widget_get_applet_locked (PanelWidget *panel,
+				GtkWidget   *applet)
+{
+	AppletData *ad;
+
+	ad = g_object_get_data (G_OBJECT (applet), PANEL_APPLET_DATA);
+	if (!ad)
+		return FALSE;
+
+	return ad->locked;
+}
+
+gboolean
+panel_widget_toggle_applet_locked (PanelWidget *panel,
+				   GtkWidget   *applet)
+{
+	AppletData *ad;
+
+	ad = g_object_get_data (G_OBJECT (applet), PANEL_APPLET_DATA);
+	if (!ad)
+		return FALSE;
+
+	return ad->locked = !ad->locked;
+}
+
 gboolean
 panel_applet_is_in_drag ()
 {
